"""Run model.

This module contains the run model.
"""

from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

from octoflow.model import value_utils

try:
    import pandas as pd
except ImportError:
    pd = None
from sqlalchemy import (
    DateTime,
    ForeignKey,
    Integer,
    String,
    Text,
)
from sqlalchemy.orm import Mapped, mapped_column

from octoflow.model.base import Base
from octoflow.model.value import Value, ValueType
from octoflow.model.variable import Variable, VariableType

__all__ = [
    "Run",
]


class Run(Base):
    """Run model."""

    __tablename__ = "run"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    experiment_id: Mapped[int] = mapped_column(Integer, ForeignKey("experiment.id", ondelete="CASCADE"))
    name: Mapped[str] = mapped_column(String, nullable=False)  # there can be multiple runs of same name
    description: Mapped[Optional[str]] = mapped_column(Text, nullable=True)
    created_at: Mapped[datetime] = mapped_column(DateTime, nullable=False, default=datetime.utcnow)
    updated_at: Mapped[datetime] = mapped_column(DateTime, nullable=False, default=datetime.utcnow)

    def _log_value(
        self,
        key: str,
        value: ValueType,
        *,
        step: Optional[Value] = None,
        type: Optional[VariableType] = None,
    ) -> Value:
        """Log a value."""
        parent_id = None if step is None else step.variable_id
        step_id = None if step is None else step.id
        with self.session():
            # get or create variable
            variable = Variable.get(
                key=key,
                experiment_id=self.experiment_id,
                parent_id=parent_id,
                type=type,
                create=True,  # create variable if it does not exist
            )
            value = Value(
                run_id=self.id,
                variable_id=variable.id,
                value=value,
                step_id=step_id,
            )
        return value

    def _log_values(
        self,
        *,
        values: Dict[str, ValueType],
        step: Optional[Value] = None,
        type: Optional[VariableType] = None,
        prefix: Optional[str] = None,
    ) -> List[Value]:
        """Log multiple values."""
        result = []
        for key, value in value_utils.flatten(values).items():
            if prefix is not None:
                key = f"{prefix}.{key}"
            value = self._log_value(
                key=key,
                value=value,
                step=step,
                type=type,
            )
            result.append(value)
        return result

    def log_param(
        self,
        key: str,
        value: ValueType,
        step: Optional[Value] = None,
    ) -> Value:
        """Log a parameter.

        Parameters
        ----------
        key : str
            Parameter key.
        value : ValueType
            Parameter value.
        step : Optional[Value], optional
            Step value, by default None.

        Returns
        -------
        Value
            Value instance.
        """
        return self._log_value(
            key=key,
            value=value,
            step=step,
            type=VariableType.parameter,
        )

    def log_params(
        self,
        values: Dict[str, ValueType],
        step: Optional[Value] = None,
        prefix: Optional[str] = None,
    ) -> List[Value]:
        """Log multiple parameters.

        Parameters
        ----------
        values : Dict[str, ValueType]
            Dictionary of values.
        step : Optional[Value], optional
            Step value, by default None.

        Returns
        -------
        List[Value]
            List of values.
        """
        return self._log_values(
            values=values,
            step=step,
            type=VariableType.parameter,
            prefix=prefix,
        )

    def log_metric(
        self,
        key: str,
        value: ValueType,
        step: Optional[Value] = None,
    ) -> Value:
        """Log a metric.

        Parameters
        ----------
        key : str
            Metric key.
        value : ValueType
            Metric value.
        step : Optional[Value], optional
            Step value, by default None.

        Returns
        -------
        Value
            Value instance.
        """
        return self._log_value(
            key=key,
            value=value,
            step=step,
            type=VariableType.metric,
        )

    def log_metrics(
        self,
        values: Dict[str, ValueType],
        step: Optional[Value] = None,
        prefix: Optional[str] = None,
    ) -> List[Value]:
        """Log multiple metrics.

        Parameters
        ----------
        values : Dict[str, ValueType]
            Dictionary of values.
        """
        return self._log_values(
            values=values,
            step=step,
            type=VariableType.metric,
            prefix=prefix,
        )

    def get_raw_logs(self) -> List[Tuple[int, int, int, str, Any]]:
        """Get all logs of the run.

        Returns
        -------
        List[Tuple[int, int, int, str, Any]]
            List of logs.
        """
        with self.session() as session:
            # get all values of this run and their corresponding variable keys
            values = (
                session.query(
                    Value.run_id,
                    Value.id,
                    Value.step_id,
                    Variable.key,
                    Value.value,
                )
                .select_from(Value)
                .join(Variable, Variable.id == Value.variable_id)
                .filter(Value.run_id == self.id)
                .distinct(Value.id)
                .all()
            )
        return values

    def get_logs(self) -> List[Dict[str, Any]]:
        """Get all logs of the run.

        Returns
        -------
        List[Dict[str, Any]]
            List of logs.
        """
        values = self.get_raw_logs()
        trees = value_utils.build_trees(values)
        return trees[self.id][None]

    def exists(self, partial: bool = True) -> bool:
        """Checks whether the run with same values for parameter typed variables exists in the database.

        Parameters
        ----------
        partial : bool, optional
            Whether to check for partial match, by default True.

        Returns
        -------
        bool
            Whether the run exists.
        """
        with self.session() as session:
            # get all values of this run and their corresponding variable keys
            values = (
                session.query(
                    Value.run_id,
                    Value.id,
                    Value.step_id,
                    Variable.key,
                    Value.value,
                )
                .select_from(Value)
                .join(Variable, Variable.id == Value.variable_id)
                .filter(
                    Variable.experiment_id == self.experiment_id,
                    Variable.type == VariableType.parameter,
                )
                .distinct(Value.id)
                .all()
            )
            # build up a tree of values
            trees = value_utils.build_trees(values)
        # check whether there is a run with same values
        if self.id not in trees:
            msg = f"run with id '{self.id}' does not exist"
            raise ValueError(msg)
        this = trees.pop(self.id)
        return any(value_utils.equals(this[None], other[None], partial=partial) for other in trees.values())



from __future__ import annotations

from typing import List, Optional

from octoflow.model.base import SessionMixin, create_engine
from octoflow.model.experiment import Experiment
from octoflow.model.run import Run


class Client(SessionMixin):
    def __init__(self, tracking_uri: Optional[str] = None) -> None:
        # if tracking_uri is None -> defaults to 'sqlite:///:memory:'
        self.tracking_uri = tracking_uri
        self.engine = create_engine(tracking_uri)

    def refresh(self) -> Client:
        self.engine = create_engine(self.tracking_uri)
        return self

    def create_experiment(
        self,
        name: str,
        description: Optional[str] = None,
        return_if_exist: bool = True,
    ) -> Experiment:
        with self.session():
            expr = self.get_experiment_by_name(name)
            if expr is not None:
                if return_if_exist:
                    return expr
                msg = f"experiment with name '{name}' already exists"
                raise ValueError(msg)
            expr = Experiment(
                name=name,
                description=description,
            )
        return expr

    def get_experiment_by_name(self, name: str) -> Experiment:
        with self.session() as session:
            expr = session.query(Experiment).filter_by(name=name).first()
        return expr

    def get_experiment(self, id: int) -> Experiment:
        if id is None:
            msg = "experiment id cannot be None"
            raise ValueError(msg)
        with self.session() as session:
            try:
                expr = session.query(Experiment).get(id)
            except ValueError:
                msg = f"experiment with id '{id}' does not exist"
                raise ValueError(msg) from None
        return expr

    def list_experiments(self) -> List[Experiment]:
        with self.session() as session:
            exprs = session.query(Experiment).all()
        return exprs

    def get_run(self, id: int) -> Run:
        with self.session() as session:
            try:
                run = session.query(Run).get(id)
            except ValueError:
                msg = f"run with id '{id}' does not exist"
                raise ValueError(msg) from None
        return run



"""Value model.

This module contains the value model.
"""

from __future__ import annotations

import json
from datetime import datetime
from typing import Dict, List, Optional, Union

from sqlalchemy import (
    JSON,
    DateTime,
    ForeignKey,
    Integer,
)
from sqlalchemy.orm import Mapped, mapped_column

from octoflow.model.base import Base

ValueType = Union[None, float, int, str, bool, List["ValueType"], Dict[str, "ValueType"]]


class Value(Base):
    """Value model."""

    __tablename__ = "value"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    run_id: Mapped[int] = mapped_column(Integer, ForeignKey("run.id", ondelete="CASCADE"))
    variable_id: Mapped[int] = mapped_column(Integer, ForeignKey("variable.id", ondelete="CASCADE"))
    value: Mapped[ValueType] = mapped_column(JSON, nullable=True)
    timestamp: Mapped[datetime] = mapped_column(DateTime, nullable=False, default=datetime.utcnow)
    step_id: Mapped[Optional[int]] = mapped_column(Integer, ForeignKey("value.id", ondelete="CASCADE"), nullable=True)

    def __repr__(self) -> str:
        """Representation of the value.

        Returns:
            str: Representation of the value.
        """
        value = json.dumps(self.value)
        if self.step_id:
            return f"Value({value}, step_id={self.step_id}, run_id={self.run_id}, variable_id={self.variable_id})"
        return f"Value({value}, run_id={self.run_id}, variable_id={self.variable_id})"



from __future__ import annotations

import math
from collections import UserList
from typing import List, Optional

import randomname
from sqlalchemy import Integer, String, Text, desc
from sqlalchemy.orm import Mapped, mapped_column

from octoflow import logging
from octoflow.model.base import Base
from octoflow.model.run import Run
from octoflow.model.variable import Variable

logger = logging.get_logger(__name__)


class Experiment(Base):
    __tablename__ = "experiment"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(String(255), nullable=False, unique=True)
    description: Mapped[str] = mapped_column(Text, nullable=True)

    def start_run(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Run:
        if name is None:
            name = randomname.get_name()
            logger.debug(f"run name generated: {name}")
        with self.session():
            run = Run(
                experiment_id=self.id,
                name=name,
                description=description,
            )
        return run

    def search_runs(
        self,
        name: Optional[str] = None,
        page: int = 1,
        per_page: Optional[int] = 10,
        # expression: Optional[Expression] = None,
    ) -> List[Run]:
        offset = (page - 1) * per_page
        with self.session() as session:
            q = session.query(Run)
            # if expression is not None:
            #     q = q.filter(Run.id.in_(expression.stmt))
            if name is not None:
                q = q.filter(Run.name == name)
            # get number of pages
            total = q.count()
            q = q.order_by(desc(Run.created_at))
            if per_page is not None:
                q = q.offset(offset).limit(per_page)
            result = q.all()
        result = UserList(result)
        result.total = total
        result.num_pages = math.ceil(total / per_page)
        return result

    def get_variables(self) -> List[Variable]:
        with self.session() as session:
            variables = session.query(Variable).filter(Variable.experiment_id == self.id).all()
        return variables



"""Variable model.

This module contains the Variable model.
"""

from __future__ import annotations

import enum
from typing import Optional

from sqlalchemy import (
    Enum,
    ForeignKey,
    Index,
    Integer,
    String,
)
from sqlalchemy.orm import Mapped, mapped_column

from octoflow.model.base import Base

__all__ = [
    "Variable",
    "VariableType",
]


class VariableType(enum.Enum):
    """Variable type."""

    unknown = 0
    parameter = 1
    metric = 2


class Variable(Base):
    __tablename__ = "variable"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    key: Mapped[str] = mapped_column(String, nullable=False)
    experiment_id: Mapped[int] = mapped_column(Integer, ForeignKey("experiment.id", ondelete="CASCADE"), nullable=False)
    parent_id: Mapped[Optional[int]] = mapped_column(
        Integer, ForeignKey("variable.id", ondelete="CASCADE"), nullable=True
    )
    type: Mapped[VariableType] = mapped_column(Enum(VariableType), nullable=False, default=VariableType.unknown)

    __table_args__ = (
        Index(
            "ix_unique_experiment_id_key",
            key,  # variables with same key should be attached to the same parent
            experiment_id,
            unique=True,
        ),
    )

    @classmethod
    def get(
        cls,
        key: str,
        experiment_id: int,
        parent_id: Optional[int] = None,
        type: VariableType = VariableType.unknown,
        create: bool = True,
    ) -> Variable:
        """Get a variable by key and experiment id.

        Parameters
        ----------
        key : str
            Variable key.
        experiment_id : int
            Experiment id.
        parent_id : Optional[int], optional
            Parent variable id, by default None.
        type : VariableType, optional
            Variable type, by default VariableType.unknown.
        create : bool, optional
            Create a new variable if not found, by default True.

        Returns
        -------
        Variable
            Variable instance.
        """
        with cls.session() as session:
            q = session.query(Variable).filter(
                Variable.key == key,
                Variable.experiment_id == experiment_id,
            )
            if parent_id is None:
                q = q.filter(Variable.parent_id.is_(None))
            else:
                q = q.filter(Variable.parent_id == parent_id)
            var = q.first()
            if var is None and create:
                # otherwise create a new variable
                var = Variable(
                    key=key,
                    experiment_id=experiment_id,
                    parent_id=parent_id,
                    type=type,
                )
        if var is None:
            return ValueError(f"unable to find variable with key '{key}'")
        return var

    def __repr__(self) -> str:
        """Override the default representation."""
        return f'Variable(experiment_id={self.experiment_id}, key="{self.key}")'

    def __eq__(self, other) -> bool:
        """Override the default Equals behavior.

        Parameters
        ----------
        other : Variable
            The Variable to compare to.

        Returns
        -------
        bool
            True if equal, False otherwise.
        """
        if isinstance(other, Variable):
            return self.experiment_id == other.experiment_id and self.key == other.key
        if isinstance(other, tuple):
            return (self.key, self.experiment_id) == other
        return False

    def __hash__(self) -> int:
        """Override the default hash behavior (that returns the id or the object).

        Returns
        -------
        int
            hash of the tuple (key, experiment_id)
        """
        return hash((self.key, self.experiment_id))



import logging
from typing import Optional, Sequence, Union

CONSOLE_HANDLER = "console"

PACKAGE_NAME = next(iter(__name__.split(".")))
DEFAULT_LOGGING_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
DEFAULT_LOGGING_HANDLERS = {
    CONSOLE_HANDLER,
}


class LoggingFactory:
    def __init__(
        self,
        name: Optional[str] = None,
        level: Union[int, str] = logging.DEBUG,
        formatter: str = DEFAULT_LOGGING_FORMAT,
        handlers: Sequence[str] = DEFAULT_LOGGING_HANDLERS,
    ) -> None:
        logger = logging.getLogger(name)
        logger.setLevel(level)
        if CONSOLE_HANDLER in handlers:
            stream_handler = logging.StreamHandler()
            stream_handler.setLevel(logging.DEBUG)
            if not isinstance(formatter, logging.Formatter):
                # Create a formatter
                formatter = logging.Formatter(formatter)
            # add it to the handlers
            stream_handler.setFormatter(formatter)
            # Add the handlers to the logger
            logger.addHandler(stream_handler)
        self.base_logger = logger

    def __call__(self, name: Optional[str] = None) -> logging.Logger:
        return logging.getLogger(name)


get_logger = LoggingFactory(PACKAGE_NAME)



from octoflow.model.base import Base
from octoflow.model.client import Client
from octoflow.model.experiment import Experiment
from octoflow.model.run import Run
from octoflow.model.value import Value

__all__ = [
    "Base",
    "Client",
    "Experiment",
    "Run",
    "Value",
]



import json
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Any, Generator, Optional, Union

import sqlalchemy as sa
from sqlalchemy import Engine, event
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import DeclarativeBase, Session, sessionmaker
from typing_extensions import Protocol, dataclass_transform, runtime_checkable  # noqa: UP035

from octoflow import logging

sessionmaker_cv = ContextVar("sessionmaker_cv", default=None)
session_cv = ContextVar("session_cv", default=None)
persist_on_init_cv = ContextVar("persist_on_init_cv", default=True)

logger = logging.get_logger(__name__)


class SessionError(ValueError): ...


@contextmanager
def persist_on_init(state: Optional[bool] = None) -> Generator[bool, None, None]:
    if state is None:
        yield persist_on_init_cv.get()
    else:
        token = persist_on_init_cv.set(state)
        yield state
        persist_on_init_cv.reset(token)


@contextmanager
def build_session(
    engine_or_session_factory: Union[Engine, sessionmaker, None] = None,
) -> Generator[Session, None, None]:
    """
    Context manager for managing SQLAlchemy sessions.

    Parameters
    ----------
    engine_or_session_factory : Engine
        SQLAlchemy Engine or SessionFactory.

    Yields
    ------
    Session
        A SQLAlchemy Session.

    Raises
    ------
    ValueError
        If the provided `engine_or_session_factory` is invalid or not in the context.
    """
    # Determine the session factory based on the input
    if isinstance(engine_or_session_factory, Engine):
        session_factory = sessionmaker(
            engine_or_session_factory,
            expire_on_commit=False,
        )
    else:
        session_factory = engine_or_session_factory
    # Set the session factory as the global session maker for this coroutine
    if session_factory is None:
        session_factory = sessionmaker_cv.get()
    sessionmaker_token = sessionmaker_cv.set(session_factory)
    # Retrieve the current session from the context
    session: Session = session_cv.get()
    if session is None:
        # No existing session found in the context
        if session_factory is None:
            msg = "no session object in the context, and session_factory is None"
            raise ValueError(msg)
        # Create a new session and set it in the context
        with session_factory() as session:
            session_token = session_cv.set(session)
            yield session  # Yield the session
            session_cv.reset(session_token)
    else:
        # Use the existing session from the context
        yield session
    # Reset the global session maker in the context
    sessionmaker_cv.reset(sessionmaker_token)


@runtime_checkable
class EngineProtocol(Protocol):
    """
    Protocol defining an object with an optional SQLAlchemy Engine.

    Attributes
    ----------
    engine : Optional[Engine]
        SQLAlchemy Engine. If not provided, it can be None.
    """

    engine: Optional[Engine]


@runtime_checkable
class SessionFactoryProtocol(Protocol):
    """
    Protocol defining an object with an optional SQLAlchemy SessionFactory.

    Attributes
    ----------
    session_factory : Optional[sessionmaker]
        SQLAlchemy SessionFactory. If not provided, it can be None.
    """

    session_factory: Optional[sessionmaker]


class SessionMixin:
    """
    Mixin class providing a context manager for managing SQLAlchemy sessions.

    Methods
    -------
    session() -> Generator[Session, None, None]
        Context manager for accessing an SQLAlchemy Session.

    Raises
    ------
    TypeError
        If the session type is unknown.
    """

    @contextmanager
    def session(self=None) -> Generator[Session, None, None]:
        """
        Context manager for managing SQLAlchemy sessions.

        Yields
        ------
        Session
            A SQLAlchemy Session.

        Raises
        ------
        TypeError
            If the protocol is unknown.
        """
        if self is None:
            # called in class context
            with build_session() as session:
                yield session
        elif isinstance(self, SessionFactoryProtocol):
            with build_session(self.session_factory) as session:
                yield session
        elif isinstance(self, EngineProtocol):
            with build_session(self.engine) as session:
                yield session
        else:
            msg = f"unknown session type: {type(self).__name__}"
            raise TypeError(msg)


@dataclass_transform()
class Base(DeclarativeBase, SessionMixin):
    def __init__(self, *args, **kwargs):
        super(Base, self).__init__()
        self.session_factory: sessionmaker = sessionmaker_cv.get()
        if self.session_factory is None:
            msg = "no session object in the context, and session_factory is None"
            raise ValueError(msg)
        n_args = len(args)
        cls = type(self)
        for idx, (key, _) in enumerate(cls.__annotations__.items()):
            if idx < n_args:
                val = args[idx]
            elif key in kwargs:
                val = kwargs[key]
            else:
                # if not provided
                continue
            super(Base, self).__setattr__(key, val)
        with persist_on_init() as persist:
            if persist:
                self.create()

    @sa.orm.reconstructor
    def init_on_load(self):
        # when object is constructed via sqlalchemy.orm
        # get a ref to the session maker
        self.session_factory: sessionmaker = sessionmaker_cv.get()

    def create(self):
        with self.session() as session:
            try:
                logger.info(f"Persisting '{type(self).__name__}'.")
                session.add(self)
                session.commit()
            except SQLAlchemyError as err:
                logger.error(f"Failed to persist '{type(self).__name__}'.")
                session.rollback()
                msg = f"unable to persist '{type(self).__name__}' object with attributes: {self.to_dict()}"
                raise ValueError(msg) from err

    def __setattr__(self, name: str, value: Any):
        if hasattr(self, "session_factory") and self.session_factory is not None:
            # update
            revert_if_fails_value = value
            with self.session() as session:
                try:
                    logger.info(f"Updating attribute '{name}' of '{type(self).__name__}' to '{value}'.")
                    super(Base, self).__setattr__(name, value)
                    session.merge(self)
                    session.commit()
                    logger.info(f"Updated attribute '{name}' of '{type(self).__name__}' to '{value}'.")
                except SQLAlchemyError as err:
                    logger.info(f"Failed to update attribute '{name}' of '{type(self).__name__}' to '{value}'.")
                    session.rollback()
                    super(Base, self).__setattr__(name, revert_if_fails_value)
                    msg = f"unable to set '{name}'"
                    raise AttributeError(msg) from err
        else:
            super(Base, self).__setattr__(name, value)

    def delete(self) -> Any:
        with self.session() as session:
            try:
                logger.info(f"Deleting '{type(self).__name__}'.")
                session.delete(self)
                session.commit()
                logger.info(f"Deleted '{type(self).__name__}'.")
            except SQLAlchemyError as err:
                logger.error(f"Failed to delete '{type(self).__name__}'.")
                session.rollback()
                msg = f"unable to delete '{type(self).__name__}'"
                raise ValueError(msg) from err

    @classmethod
    def from_dict(cls, data: dict):
        primary_keys = iter(column.name for column in cls.__table__.columns if column.primary_key)
        kwargs = {key: data[key] for key in primary_keys}
        session_factory: sessionmaker = sessionmaker_cv.get()
        with build_session(session_factory) as session:
            # merge with the current session
            self = session.query(cls).filter_by(**kwargs).first()
        return self

    def to_dict(self) -> dict[str, Any]:
        return {key: getattr(self, key) for key in type(self).__annotations__}

    def to_json(self):
        data = self.to_dict()
        return json.dumps(data, sort_keys=True, default=str)


def create_engine(url: Optional[str] = None):
    """
    Create a SQLAlchemy engine with an optional database URL.

    Parameters
    ----------
    url : Optional[str], optional
        Database URL, by default None.
        If None, an SQLite in-memory database is created for testing purposes.

    Returns
    -------
    Engine
        SQLAlchemy Engine instance.

    Notes
    -----
    This function also creates tables defined in the declarative base (Base).

    Examples
    --------
    >>> engine = create_engine("sqlite:///logs/octoflow/database.sqlite")
    >>> engine = create_engine("postgresql://user:password@localhost/mydatabase")
    >>> engine = create_engine()  # Creates an SQLite in-memory database for testing
    """
    if url is None:
        # Create an SQLite in-memory database for testing purposes
        url = "sqlite:///:memory:"
    # Create the SQLAlchemy engine
    engine = sa.create_engine(url, echo=False)
    event.listen(engine, "connect", _fk_pragma_on_connect)
    # Create tables defined in the declarative base
    Base.metadata.create_all(engine)
    return engine


def _fk_pragma_on_connect(dbapi_con, con_record):
    dbapi_con.execute("pragma foreign_keys=ON")



from collections import UserDict
from collections.abc import Mapping
from typing import Any, Dict, List


def flatten(
    data: Dict[str, Any],
    parent_key: str = "",
    separator: str = ".",
) -> dict[str, Any]:
    items = []
    for key, value in data.items():
        # escape dots
        new_key = parent_key + separator + key if parent_key else key
        if isinstance(value, Mapping):
            items.extend(
                flatten(
                    value,
                    parent_key=new_key,
                    separator=separator,
                ).items()
            )
        else:
            items.append((new_key, value))
    return dict(items)


def build_trees_base(values: List) -> Dict[int, dict]:
    """
    Build a value tree from a list of values.

    Parameters
    ----------
    values : List
        A list of values containing tuples of (run_id, value_id, step_id, key, value).

    Returns
    -------
    Dict[int, dict]
        A dictionary representing the value tree, where the keys are run IDs and the values are nested dictionaries.
    """
    trees: dict = {}
    id2value = {}
    for run_id, value_id, step_id, key, value in values:
        id2value[value_id] = (key, value)
        if run_id not in trees:
            trees[run_id] = {}
        root: dict = trees[run_id]
        if step_id not in root or root[step_id] is None:
            root[step_id] = {}
        node = root[value_id] if value_id in root else root.setdefault(value_id, {})
        root[step_id][value_id] = node
    trees = UserDict(trees)
    trees.id2value = id2value
    return trees


def build_trees(values: List) -> Dict[int, dict]:
    """
    Build a value tree from a list of values.

    Parameters
    ----------
    values : List
        A list of values containing tuples of (run_id, value_id, step_id, key, value).

    Returns
    -------
    Dict[int, dict]
        A dictionary representing the value tree, where the keys are run IDs and the values are nested dictionaries.
    """
    trees = build_trees_base(values)
    # flatten the tree
    run_tree_map = {}
    for run_id, tree in trees.items():
        run_tree_map[run_id] = _normalize(tree, trees.id2value)
    return run_tree_map


def _normalize(tree: dict, id2value: dict) -> dict:
    """
    Normalize a value tree by removing IDs and replacing them with key-value pairs.

    Parameters
    ----------
    tree : dict
        The value tree to be normalized.
    id2value : dict
        A dictionary mapping value IDs to their corresponding key-value pairs.

    Returns
    -------
    dict
        The normalized value tree.
    """
    nominal_tree = {}
    for parent_id, child_id in tree.items():
        parent_value = id2value.get(parent_id)
        if isinstance(child_id, dict):
            nominal_tree[parent_value] = _normalize(child_id, id2value)
        else:
            child_value = id2value[child_id]
            nominal_tree[parent_value] = child_value
    return nominal_tree


def equals(
    p: dict,
    q: dict,
    /,
    partial: bool = True,
) -> bool:
    """
    Check whether two dictionaries are equal.

    Parameters
    ----------
    p : dict
        The first dictionary to compare.
    q : dict
        The second dictionary to compare.
    partial : bool, optional
        If True, performs a partial comparison where missing keys in one dictionary are ignored (default is True).

    Returns
    -------
    bool
        True if the dictionaries are equal, False otherwise.

    Raises
    ------
    ValueError
        If either `p` or `q` is not a dictionary.
    """
    if not partial:
        if p.keys() != q.keys():
            return False
        return all(equals(value, q[key]) for key, value in p.items())
    if not isinstance(p, dict) or not isinstance(q, dict):
        msg = f"cannot compare {type(p).__name__} and {type(q).__name__} types"
        raise ValueError(msg)
    for key, val_a in p.items():
        if key not in q:
            return False
        val_b = q[key]
        if not equals(val_a, val_b):
            return False
    return True
